# $Id: disco.tcl,v 1.37 2005/08/06 22:13:32 aleksey Exp $

set ::NS(disco_items) "http://jabber.org/protocol/disco#items"
set ::NS(disco_info)  "http://jabber.org/protocol/disco#info"

option add *JDisco.fill          Black		widgetDefault
option add *JDisco.featurecolor  MidnightBlue   widgetDefault
option add *JDisco.identitycolor DarkGreen      widgetDefault
option add *JDisco.optioncolor   DarkViolet     widgetDefault

namespace eval disco {
    variable supported_nodes
}

proc disco::request_items {jid node args} {
    variable disco

    set handler {}
    set cache no

    foreach {attr val} $args {
	switch -- $attr {
	    -handler    {set handler $val}
	    -cache      {set cache $val}
	    -connection {set connid $val}
	}
    }
    if {![info exists connid]} {
	set connid [jlib::route $jid]
    }

    switch -- $cache {
	first -
	only -
	yes {
	    if {[info exists disco(items,$jid,$node)]} {
		set items $disco(items,$jid,$node)
		if {$handler != ""} {
		    eval $handler [list OK $items]
		}
		if {$cache != "first"} {
		    return [list OK $items]
		}
	    } elseif {$cache == "only"} {
		return NO
	    }
	}
    }

    set vars [list xmlns $::NS(disco_items)]
    if {$node != ""} {
	lappend vars node $node
    }

    jlib::send_iq get \
	    [jlib::wrapper:createtag query \
		 -vars $vars] \
	-to $jid \
	-connection $connid \
	-command [list [namespace current]::parse_items $jid $node $handler]
}

proc disco::parse_items {jid node handler res child} {
    variable disco

    if {![cequal $res OK]} {
	if {$handler != ""} {
	    eval $handler [list ERR $child]
	}
	hook::run disco_items_hook $jid $node ERR $child
	return
    }

    set items {}

    jlib::wrapper:splitxml $child tag vars isempty chdata childrens

    foreach ch $childrens {
	jlib::wrapper:splitxml $ch tag1 vars1 isempty1 chdata1 childrens1
	switch -- $tag1 {
	    item {
		set ijid  [jlib::wrapper:getattr $vars1 jid]
		set inode  [jlib::wrapper:getattr $vars1 node]
		set name  [jlib::wrapper:getattr $vars1 name]
		lappend items [list jid $ijid node $inode name $name]
		set disco(jidname,$ijid,$inode) $name
	    }
	}
    }

    set disco(items,$jid,$node) $items

    debugmsg disco "ITEMS: [list $items]"

    if {$handler != ""} {
	eval $handler [list OK $items]
    }

    hook::run disco_items_hook $jid $node OK $items
}



proc disco::request_info {jid node args} {
    variable disco

    set handler {}
    set cache no

    foreach {attr val} $args {
	switch -- $attr {
	    -handler    {set handler $val}
	    -cache      {set cache $val}
	    -connection {set connid $val}
	}
    }
    if {![info exists connid]} {
	set connid [jlib::route $jid]
    }

    switch -- $cache {
	first -
	only -
	yes {
	    if {[info exists disco(info,identities,$jid,$node)] && \
		    [info exists disco(info,identities,$jid,$node)]} {
		set identities $disco(info,identities,$jid,$node)
		set features   $disco(info,features,$jid,$node)
		set extras     $disco(info,extras,$jid,$node)
		if {$handler != ""} {
		    eval $handler [list OK $identities $features $extras]
		}
		if {$cache != "first"} {
		    return [list OK $identities $features $extras]
		}
	    } elseif {$cache == "only"} {
		return NO
	    }
	}
    }

    set vars [list xmlns $::NS(disco_info)]
    if {$node != ""} {
	lappend vars node $node
    }

    jlib::send_iq get \
	[jlib::wrapper:createtag query \
	     -vars $vars] \
	-to $jid \
	-connection $connid \
	-command [list [namespace current]::parse_info $jid $node $handler]
}

proc disco::parse_info {jid node handler res child} {
    variable disco

    if {![cequal $res OK]} {
	if {$handler != ""} {
	    eval $handler [list ERR $child {} {}]
	}
	hook::run disco_info_hook $jid $node ERR $child {} {}
	return
    }

    set identities {}
    set features {}
    set extras {}

    jlib::wrapper:splitxml $child tag vars isempty chdata children

    foreach ch $children {
	jlib::wrapper:splitxml $ch tag1 vars1 isempty1 chdata1 children1
	switch -- $tag1 {
	    identity {
		lappend identities \
		    [list \
			 category [jlib::wrapper:getattr $vars1 category] \
			 name [jlib::wrapper:getattr $vars1 name] \
			 type [jlib::wrapper:getattr $vars1 type]]
	    }
	    feature {
		set var [jlib::wrapper:getattr $vars1 var]
		if {$var == ""} {
		    set var [jlib::wrapper:getattr $vars1 type]
		}
		lappend features [list var $var]
	    }
	    default {
		if {[jlib::wrapper:getattr $vars1 xmlns] == "jabber:x:data" && \
			[jlib::wrapper:getattr $vars1 type] == "result"} {
		    set extras [data::parse_xdata_results $children1]
		}
	    }
	}
    }

    set disco(info,identities,$jid,$node) $identities
    set disco(info,features,$jid,$node) $features
    set disco(info,extras,$jid,$node) $extras

    debugmsg disco \
	"INFO: IDENTITIES [list $identities] FEATURES [list $features]\
EXTRAS [list $extras]"

    if {$handler != ""} {
	eval $handler [list OK $identities $features $extras]
    }
    hook::run disco_info_hook $jid $node OK $identities $features $extras
}

proc disco::get_jid_name {jid node} {
    variable disco
    if {[info exists disco(jidname,$jid,$node)]} {
	return $disco(jidname,$jid,$node)
    } else {
	return ""
    }
}

proc disco::get_jid_identities {jid node} {
    variable disco
    if {[info exists disco(info,identities,$jid,$node)]} {
	return $disco(info,identities,$jid,$node)
    } else {
	return {}
    }
}

###############################################################################

proc disco::info_query_get_handler {connid from child} {
    variable supported_nodes
    variable node_handlers

    jlib::wrapper:splitxml $child tag vars isempty chdata children
    set node [jlib::wrapper:getattr $vars node]

    if {![cequal $node ""]} {
	if {![info exists supported_nodes($node)]} {
	    return {error cancel not-allowed}
	} else {
	    set restags [$node_handlers($node) info $connid $from $child]
	    if {[cequal [lindex $restags 0] error]} {
		return $restags
	    } else {
		set res [jlib::wrapper:createtag query \
			     -vars [list xmlns $::NS(disco_info) node $node] \
			     -subtags $restags]
	    }
	}
    } else {
	set restags {}

	lappend restags [jlib::wrapper:createtag identity \
			     -vars [list category client \
					 type pc \
					 name Tkabber]]

	foreach ns $::iq::supported_ns {
	    lappend restags [jlib::wrapper:createtag feature \
				 -vars [list var $ns]]
	}
    
	set res [jlib::wrapper:createtag query \
		     -vars [list xmlns $::NS(disco_info)] \
		     -subtags $restags]
    }
    return [list result $res]
}

iq::register_handler get query $::NS(disco_info) \
    [namespace current]::disco::info_query_get_handler

proc disco::register_node {node handler {name ""}} {
    variable supported_nodes
    variable node_handlers

    set supported_nodes($node) $name
    set node_handlers($node) $handler
}

proc disco::items_query_get_handler {connid from child} {
    variable supported_nodes
    variable node_handlers

    jlib::wrapper:splitxml $child tag vars isempty chdata children
    set node [jlib::wrapper:getattr $vars node]

    if {![cequal $node ""]} {
	if {![info exists supported_nodes($node)]} {
	    return {error cancel not-allowed}
	} else {
	    set restags [$node_handlers($node) items $connid $from $child]
	    if {[cequal [lindex $restags 0] error]} {
		return $restags
	    } else {
		set res [jlib::wrapper:createtag query \
			     -vars [list xmlns $::NS(disco_items) node $node] \
			     -subtags $restags]
	    }
	}
    } else {
	set restags {}

	set bare_from [node_and_server_from_jid $from]
	set chatid [chat::chatid $connid $bare_from]
	if {[chat::is_groupchat $chatid]} {
	    set myjid [chat::our_jid $chatid]
	} else {
	    set myjid [jlib::connection_jid $connid]
	}

	foreach node [array names supported_nodes] {
	    set vars [list jid $myjid]
	    if {![cequal $supported_nodes($node) ""]} {
		lappend vars name $supported_nodes($node)
	    }
	    if {![cequal $node ""]} {
		lappend vars node $node
	    }
	    lappend restags [jlib::wrapper:createtag item \
				 -vars $vars]
	}

	set res [jlib::wrapper:createtag query \
		     -vars [list xmlns $::NS(disco_items)] \
		     -subtags $restags]
    }
    return [list result $res]
}

iq::register_handler get query $::NS(disco_items) \
    [namespace current]::disco::items_query_get_handler

###############################################################################
# Disco Browser

namespace eval disco::browser {
    set winid 0

    set icon(group)   [Bitmap::get [pixmap browser conference.gif]]
    set icon(client)  [Bitmap::get [pixmap browser user.gif]]
    set icon(jud)     [Bitmap::get [pixmap browser jud.gif]]
    set icon(aim)     [Bitmap::get [pixmap browser aim.gif]]
    set icon(icq)     [Bitmap::get [pixmap browser icq.gif]]
    set icon(msn)     [Bitmap::get [pixmap browser msn.gif]]
    set icon(gg)      [Bitmap::get [pixmap browser gg.gif]]
    set icon(weather) [Bitmap::get [pixmap browser weather.gif]]
    set icon(yahoo)   [Bitmap::get [pixmap browser yahoo.gif]]
    set icon(message) [Bitmap::get [pixmap browser message.gif]]
    image create photo ""

    custom::defvar disco_list {} [::msgcat::mc "List of discovered JIDs."] \
	    -group Hidden
    custom::defvar node_list {} [::msgcat::mc "List of discovered JID nodes."] \
	    -group Hidden
}

proc disco::browser::open_win {jid args} {
    variable winid
    variable disco
    variable config
    variable curjid
    variable disco_list
    variable node_list
    variable browser

    if {[llength [jlib::connections]] == 0} return

    foreach {opt val} $args {
	switch -- $opt {
	    -connection { set connid $val }
	}
    }
    if {![info exists connid]} {
	set connid [jlib::route $jid]
    }

    if {$jid == ""} {
	set curjid($winid) [jlib::connection_server $connid]
    } else {
	set curjid($winid) $jid
    }

    set w .disco_$winid
    set wid $winid
    incr winid
    set browser(connid,$w) $connid

    add_win $w -title [::msgcat::mc "Service Discovery"] \
	-tabtitle [::msgcat::mc "Discovery"] \
	-raisecmd [list focus $w.tree] \
	-class JDisco

    set config(fill) 	      [option get $w fill          JDisco]
    set config(featurecolor)  [option get $w featurecolor  JDisco]
    set config(identitycolor) [option get $w identitycolor JDisco]
    set config(optioncolor)   [option get $w optioncolor   JDisco]

    frame $w.navigate
    button $w.navigate.back -text <- \
	-command [list [namespace current]::history_move $w 1]
    button $w.navigate.forward -text -> \
	-command [list [namespace current]::history_move $w -1]
    label $w.navigate.lentry -text [::msgcat::mc "JID:"]
    ComboBox $w.navigate.entry -textvariable [namespace current]::curjid($wid) \
	-dropenabled 1 -droptypes {JID {}} \
	-dropcmd [list [namespace current]::entrydropcmd $w] \
	-command [list [namespace current]::go $w] \
	-values $disco_list
    label $w.navigate.lnode -text [::msgcat::mc "Node:"]
    ComboBox $w.navigate.node -textvariable [namespace current]::curnode($wid) \
	-values $node_list -width 20
    button $w.navigate.browse -text [::msgcat::mc "Browse"] \
	-command [list [namespace current]::go $w]

    #bind $w.navigate.entry <Return> [list disco::go $w]

    pack $w.navigate.back $w.navigate.forward $w.navigate.lentry -side left
    pack $w.navigate.browse -side right
    pack $w.navigate.entry -side left -expand yes -fill x
    pack $w.navigate.lnode -side left
    pack $w.navigate.node -side left -expand no -fill x
    pack $w.navigate -fill x


    set sw [ScrolledWindow $w.sw]

    set tw [Tree $w.tree -deltax 16 -deltay 18 -dragenabled 1 \
		-draginitcmd [namespace current]::draginitcmd]
    $sw setwidget $tw

    pack $sw -side top -expand yes -fill both
    set disco(tree,$w) $tw
    $tw bindText <Double-ButtonPress-1> \
	[list [namespace current]::textaction $w]
    $tw bindText <ButtonPress-3> \
	[list [namespace current]::textpopup $w]
    $tw bindText <Any-Enter>  \
	[list [namespace current]::textballoon $w enter  %X %Y]
    $tw bindText <Any-Motion> \
	[list [namespace current]::textballoon $w motion %X %Y]
    $tw bindText <Any-Leave>  \
	[list [namespace current]::textballoon $w leave  %X %Y]
    bindscroll $tw.c

    # HACK
    bind $tw.c <Return> \
	"[namespace current]::textaction $w \[$tw selection get\]"

    lappend browser(opened) $w
    set browser(opened) [lrmdups $browser(opened)]
    set browser(required,$w) {}
    set browser(tree,$w) $tw

    set browser(hist,$w) {}
    set browser(histpos,$w) 0

    go $w
}

proc disco::browser::go {bw} {
    variable browser
    variable disco_list
    variable node_list
    
    if {[winfo exists $bw]} {
	set jid [$bw.navigate.entry.e get]
	set node [$bw.navigate.node.e get]

	history_add $bw [list $jid $node]

        set disco_list [update_combo_list $disco_list $jid 20]
	$bw.navigate.entry configure -values $disco_list
	set custom::saved([namespace current]::disco_list) $disco_list
        set node_list [update_combo_list $node_list $node 20]
	$bw.navigate.node configure -values $node_list
	set custom::saved([namespace current]::node_list) $node_list
	custom::store

	lappend browser(required,$bw) $jid
	set browser(required,$bw) [lrmdups $browser(required,$bw)]

	disco::request_items $jid $node -connection $browser(connid,$bw)
	disco::request_info $jid $node -connection $browser(connid,$bw)
    }
}

proc disco::browser::info_receive {jid node res identities features extras} {
    variable browser

    if {![info exists browser(opened)]} return

    foreach w $browser(opened) {
	if {[winfo exists $w] && [lcontain $browser(required,$w) $jid]} {
	    draw_info $w $jid $node $res $identities $features $extras
	}
    }
}

hook::add disco_info_hook \
    [namespace current]::disco::browser::info_receive

proc disco::browser::draw_info {w jid node res identities features extras} {
    variable browser
    variable config
    global font

    set tw $browser(tree,$w)

    set name [disco::get_jid_name $jid $node]
    set tnode [jid_to_tag [list $jid $node]]
    set parent_tag [jid_to_tag [list $jid $node]]
    set data [list item $jid $node]
    set desc [item_desc $jid $node $name]
    set icon ""

    add_line $tw $parent_tag $tnode $icon $desc $data -font $font \
	-fill $config(fill)

    if {$res != "OK"} {
	set tnode [jid_to_tag "error info $jid $node"]
	set data [list error_info $jid]
	#set name     [jlib::wrapper:getattr $identity name]
	set desc [format [::msgcat::mc "Error getting info: %s"] \
		      [error_to_string $identities]]
	set icon ""
	
	add_line $tw $parent_tag $tnode $icon $desc $data -font $font \
	    -fill $config(identitycolor)

	remove_old $tw $parent_tag identity [list $tnode]
	remove_old $tw $parent_tag feature [list $tnode]
	reorder_node $tw $parent_tag
	return
    }


    set identitynodes {}

    set category ""
    set type ""
    foreach identity $identities {
	set tnode [jid_to_tag "identity $identity $jid $node"]
	lappend identitynodes $tnode
	set data [list identity $jid $node]
	set name     [jlib::wrapper:getattr $identity name]
	set category [jlib::wrapper:getattr $identity category]
	set type     [jlib::wrapper:getattr $identity type]
	set desc "$name ($category/$type)"
	set icon [item_icon $category $type]
	
	add_line $tw $parent_tag $tnode $icon $desc $data -font $font \
	    -fill $config(identitycolor)
    }

    set extranodes {}
    
    foreach extra $extras {
	lassign $extra var label value
	set tnode [jid_to_tag "extra $var $jid $node"]
	lappend extranodes $tnode
	set data [list extra $var $jid $node]
	set desc "$label ($var): $value"
	set icon ""
	
	add_line $tw $parent_tag $tnode $icon $desc $data -font $font \
	    -fill $config(identitycolor)
    }

    set featurenodes {}

    foreach feature $features {
	set var [jlib::wrapper:getattr $feature var]
	set tnode [jid_to_tag "feature $feature $jid $node"]
	lappend featurenodes $tnode
	set data [list feature $jid $node $feature $category $type]
	set desc $var
	if {[info exists browser(feature_handler_desc,$var)]} {
	    catch { array unset tmp }
	    array set tmp $browser(feature_handler_desc,$var)
	    if {[info exists tmp($category)]} {
		set desc "$tmp($category) ($var)"
	    } elseif {[info exists tmp(*)]} {
		set desc "$tmp(*) ($var)"
	    }
	}
	# TODO
	set icon ""; #[item_icon $category $type]

	add_line $tw $parent_tag $tnode $icon $desc $data \
	    -fill $config(featurecolor)
    }

    remove_old $tw $parent_tag identity $identitynodes
    remove_old $tw $parent_tag extra    $extranodes
    remove_old $tw $parent_tag feature  $featurenodes
    remove_old $tw $parent_tag error_info {}
    reorder_node $tw $parent_tag
}

proc disco::browser::items_receive {jid node res items} {
    variable browser

    if {![info exists browser(opened)]} return

    foreach w $browser(opened) {
	if {[winfo exists $w] && [lcontain $browser(required,$w) $jid]} {
	    draw_items $w $jid $node $res $items
	}
    }
}

hook::add disco_items_hook \
    [namespace current]::disco::browser::items_receive

proc disco::browser::draw_items {w jid node res items} {
    variable browser
    variable config
    global font

    set tw $browser(tree,$w)

    set tnode [jid_to_tag [list $jid $node]]
    set parent_tag [jid_to_tag [list $jid $node]]
    if {![info exists browser(sort,$w,$parent_tag)]} {
	set browser(sort,$w,$parent_tag) sort
    }
    set data [list item $jid $node]
    set desc $jid
    set icon ""

    add_line $tw $parent_tag $tnode $icon $desc $data -font $font \
	-fill $config(fill)

    if {$res != "OK"} {
	set tnode [jid_to_tag "error items $jid $node"]
	set data [list error_items $jid]
	#set name     [jlib::wrapper:getattr $identity name]
	set desc [format [::msgcat::mc "Error getting items: %s"] \
		      [error_to_string $items]]
	set icon ""
	
	add_line $tw $parent_tag $tnode $icon $desc $data -font $font \
	    -fill $config(fill)

	remove_old $tw $parent_tag item [list $tnode]
	reorder_node $tw $parent_tag
	return
    }

    set itemnodes {}

    foreach item $items {
	set ijid [jlib::wrapper:getattr $item jid]
	set node [jlib::wrapper:getattr $item node]
	set name [jlib::wrapper:getattr $item name]
	set tnode [jid_to_tag [list $ijid $node]]
	lappend itemnodes $tnode
	set data [list item $ijid $node]
	set desc [item_desc $ijid $node $name]
	set icon ""

	add_line $tw $parent_tag $tnode $icon $desc $data -font $font \
	    -fill $config(fill)
    }
    remove_old $tw $parent_tag item $itemnodes
    remove_old $tw $parent_tag error_items {}
    #reorder_node $tw $parent_tag
    browser_action $browser(sort,$w,$parent_tag) $w $parent_tag
}

proc disco::browser::negotiate_feature {tw connid jid parent type} {
    variable config
    global font

    lassign [negotiate::send_request $connid $jid $type] res opts

    if {![winfo exists $tw]} return

    if {$res != "OK"} {
	set node [jid_to_tag "error negotiate $parent"]
	set data [list error_negotiate $parent $jid]
	#set name     [jlib::wrapper:getattr $identity name]
	set desc [format [::msgcat::mc "Error negotiate: %s"] \
		      [error_to_string $opts]]
	set icon ""
	
	add_line $tw $parent $node $icon $desc $data -font $font \
	    -fill $config(optioncolor)

	remove_old $tw $parent option [list $node]
	return
    }

    set optnodes {}

    foreach opt $opts {
	set node [jid_to_tag "option $opt $parent"]
	lappend optnodes $node
	set data [list option $opt $node]
	set desc $opt
	set icon ""
	
	add_line $tw $parent $node $icon $desc $data -font $font \
	    -fill $config(optioncolor)
    }
    remove_old $tw $parent option $optnodes
    remove_old $tw $parent error_negotiate {}
}


proc disco::browser::add_line {tw parent node icon desc data args} {

    if {[$tw exists $node]} {
	if {[$tw parent $node] != $parent && [$tw exists $parent] && \
		$parent != $node} {
	    if {[catch { $tw move $parent $node end }]} {
		debugmsg disco "MOVE FAILED: $parent $node"
	    } else {
		debugmsg disco "MOVE: $parent $node"
	    }
	}
	if {[$tw itemcget $node -data] != $data || \
		[$tw itemcget $node -text] != $desc} {
	    debugmsg disco RECONF
	    $tw itemconfigure $node -text $desc -data $data
	}
    } elseif {[$tw exists $parent]} {
	eval {$tw insert end $parent $node -text $desc -open 1 -image $icon \
		  -data $data} $args
    } else {
	eval {$tw insert end root $node -text $desc -open 1 -image $icon \
		  -data $data} $args
    }

}


proc disco::browser::reorder_node {tw node {order {}}} {
    set subnodes [$tw nodes $node]

    set identities {}
    set features {}
    set items {}
    foreach sn $subnodes {
	lassign [$tw itemcget $sn -data] kind
	switch -- $kind {
	    error_items -
	    item     {lappend items      $sn}
	    error_info -
	    identity {lappend identities $sn}
	    feature  {lappend features   $sn}
	}
    }
    if {$order == {}} {
	$tw reorder $node [concat $identities $features $items]
    } else {
	$tw reorder $node [concat $identities $features $order]
    }
}

proc disco::browser::remove_old {tw node kind newnodes} {
    set subnodes [$tw nodes $node]

    set items {}
    foreach sn $subnodes {
	lassign [$tw itemcget $sn -data] kind1
	if {$kind == $kind1 && ![lcontain $newnodes $sn]} {
	    $tw delete $sn
	}
    }
}

proc disco::browser::item_desc {jid node name} {
    if {$node != ""} {
	set snode " \[$node\]"
    } else {
	set snode ""
    }
    if {![cequal $name ""]} {
	return "$name$snode ($jid)"
    } else {
	return $jid$snode
    }
}

proc disco::browser::item_icon {category type} {
    variable icon
    switch -- $category {
	gateway {
	    switch -- $type {
		aim {return $icon(aim)}
		icq {return $icon(icq)}
		msn {return $icon(msn)}
		yahoo {return $icon(yahoo)}
		x-gadugadu {return $icon(gg)}
		default {return ""}
	    }
	}
	service {
	    switch -- $type {
		jud {return $icon(jud)}
		aim {return $icon(aim)}
		icq {return $icon(icq)}
		msn {return $icon(msn)}
		x-gadugadu {return $icon(gg)}
		x-weather {return $icon(weather)}
		yahoo {return $icon(yahoo)}
		default {return ""}
	    }
	}
	application {
	    switch -- $type {
		x-weather {return $icon(weather)}
		default {return ""}
	    }
	}
	conference {
	    return $icon(group)
	}
	user -
	client {
	    return $icon(client)
	}
	directory {
	    return $icon(jud)
	}
	headline {
	    return $icon(message)
	}
	default {
	    return ""
	}
    }
}

proc disco::browser::textaction {bw tnode} {
    variable disco
    variable browser

    set tw $browser(tree,$bw)
    set data [$tw itemcget $tnode -data]
    set data2 [lassign $data type]
    switch -- $type {
	item {
	    lassign $data2 jid node
	    goto $bw $jid $node
	}
	feature {
	    lassign $data2 jid node feature category subtype
	    set var [jlib::wrapper:getattr $feature var]
	    debugmsg disco $jid
	    if {$var != ""} {
		if {[info exists browser(feature_handler,$var)]} {
		    if {$browser(feature_handler_node,$var)} {
			eval $browser(feature_handler,$var) [list $jid $node \
			    -category $category -type $subtype \
			    -connection $browser(connid,$bw)]
		    } else {
			eval $browser(feature_handler,$var) [list $jid \
			    -category $category -type $subtype \
			    -connection $browser(connid,$bw)]
		    }
		} else {
		    negotiate_feature $tw $browser(connid,$bw) $jid $tnode $var
		}
	    }
	}
    }
}

proc disco::browser::textpopup {bw tnode} {
    variable browser

    if {[winfo exists [set m .discopopupmenu]]} {
	destroy $m
    }
    menu $m -tearoff 0

    set tw $browser(tree,$bw)
    set data [$tw itemcget $tnode -data]
    set data2 [lassign $data type]

    switch -- $type {
	identity {
	    return
	}

	feature {
	    $m add command -label [::msgcat::mc "Browse"] \
		-command [list [namespace current]::browser_action browse $bw $tnode]
	}

	item {
	    lassign $data2 jid node
	    set identities [[namespace parent]::get_jid_identities $jid $node]
	    # TODO: use all identities
	    set category [jlib::wrapper:getattr [lindex $identities 0] category]
	    switch -- $category {
		client {
		    message::subject_menu $m \
			   [lindex $data 1] message -connection $browser(connid,$bw)
		    tk_popup $m [winfo pointerx .] [winfo pointery .]
		    return
		}

		conference {
		    if {[string first @ [set headjid [lindex $data 1]]] > 0} {
			$m add command -label [::msgcat::mc "Join group..."] \
			    -command [list join_group_dialog \
					   -server [server_from_jid $headjid] \
					   -group [node_from_jid $headjid] \
					   -connection $browser(connid,$bw)]
			$m add command -label [::msgcat::mc "Add conference..."] \
			    -command [list plugins::conferences::add_conference_dialog \
					   -group [node_from_jid $headjid] \
					   -server [server_from_jid $headjid] \
					   -connection $browser(connid,$bw)]
			$m add separator
		    }
		}

		default {
		}
	    }
	    $m add command -label [::msgcat::mc "Browse"] \
		-command [list [namespace current]::browser_action browse $bw $tnode]
	    $m add command -label [::msgcat::mc "Sort items by name"] \
		-command [list [namespace current]::browser_action sort $bw $tnode]
	    $m add command -label [::msgcat::mc "Sort items by JID/node"] \
		-command [list [namespace current]::browser_action sortjid $bw $tnode]
	}
    }

    tk_popup $m [winfo pointerx .] [winfo pointery .]
}

proc disco::browser::browser_action {action bw tnode} {
    variable browser

    set tw $browser(tree,$bw)
    set data [$tw itemcget $tnode -data]
    set data2 [lassign $data type]

    switch -glob -- $type/$action {
	item/browse -
	feature/browse {
	    textaction $bw $tnode
	}

	item/sort {
	    set browser(sort,$bw,$tnode) sort
            set items {}
            foreach child [$tw nodes $tnode] {
		set data [lassign [$tw itemcget $child -data] type]
		switch -- $type {
		    item {
			lassign $data jid node
			lappend items \
			    [list $child \
				  [[namespace parent]::get_jid_name $jid $node]]
		    }
		}
            }
            set neworder {}
            foreach item [lsort -dictionary -index 1 $items] {
                lappend neworder [lindex $item 0]
            }
            reorder_node $tw $tnode $neworder

            foreach child [$tw nodes $tnode] {
                browser::browser_action $action $bw $child
            }
	}

	item/sortjid {
	    set browser(sort,$bw,$tnode) sortjid
            set items {}
	    set items_with_nodes {}
            foreach child [$tw nodes $tnode] {
		set data [lassign [$tw itemcget $child -data] type]
		switch -- $type {
		    item {
			lassign $data jid node
			if {$node != {}} {
			    lappend items_with_nodes \
				[list $child [join $data "\u0000"]]
			} else {
			    lappend items [list $child $jid]
			}
		    }
		}
            }
            set neworder {}
            foreach item [concat [lsort -dictionary -index 1 $items] \
				 [lsort -dictionary -index 1 $items_with_nodes]] {
                lappend neworder [lindex $item 0]
            }
            reorder_node $tw $tnode $neworder

            foreach child [$tw nodes $tnode] {
                browser::browser_action $action $bw $child
            }
	}

	default {
	}
    }
}

# TODO
proc disco::browser::textballoon {bw action X Y node} {
    variable disco
    variable browser

    set tw $browser(tree,$bw)
    set data [$tw itemcget $node -data]
    set data2 [lassign $data type]
    if {$type == "jid"} {
	    lassign $data2 jid category type name version
    } else {
	return
    }

    switch -- $action {
	enter {
	    balloon::set_text \
		[item_balloon_text $jid $category $type $name $version]
	}
	motion {
	    balloon::on_mouse_move \
		[item_balloon_text $jid $category $type $name $version] \
		$X $Y
	}
	leave {
	    balloon::destroy
	}
    }
}

proc disco::browser::goto {bw jid node} {
    $bw.navigate.entry.e delete 0 end
    $bw.navigate.entry.e insert 0 $jid
    $bw.navigate.node.e delete 0 end
    $bw.navigate.node.e insert 0 $node
    go $bw
}

proc disco::browser::draginitcmd {t tnode top} {
    set data [$t itemcget $tnode -data]
    set data2 [lassign $data type jid node]

    if {$type == "item"} {
	if {[set img [$t itemcget $tnode -image]] != ""} {
	    pack [label $top.l -image $img -padx 0 -pady 0]
	}
	
	return [list JID {copy} [list $jid "" "" "" ""]]
    } else {
	return {}
    }
}

proc disco::browser::entrydropcmd {bw target source pos op type data} {
    set jid [lindex $data 0]
    goto $bw $jid ""
}

proc disco::browser::history_move {bw shift} {
    variable browser

    set newpos [expr {$browser(histpos,$bw) + $shift}]

    if {$newpos < 0} {
	return
    }

    if {$newpos >= [llength $browser(hist,$bw)]} {
	return
    }

    set newjidnode [lindex $browser(hist,$bw) $newpos]
    set browser(histpos,$bw) $newpos

    lassign $newjidnode newjid newnode
    $bw.navigate.entry.e delete 0 end
    $bw.navigate.entry.e insert 0 $newjid
    $bw.navigate.node.e delete 0 end
    $bw.navigate.node.e insert 0 $newnode

    disco::request_items $newjid $newnode -connection $browser(connid,$bw)
    disco::request_info $newjid $newnode -connection $browser(connid,$bw)
}

proc disco::browser::history_add {bw jid} {
    variable browser

    set browser(hist,$bw) [lreplace $browser(hist,$bw) 0 \
			       [expr {$browser(histpos,$bw) - 1}]]
    
    lvarpush browser(hist,$bw) $jid
    set browser(histpos,$bw) 0
    debugmsg disco $browser(hist,$bw)
}

#proc disco::browser::item_balloon_text {jid category type name version} {
#    variable disco
#    set text [format [::msgcat::mc "%s: %s/%s, Description: %s, Version: %s\nNumber of children: %s"] \
#	    $jid $category $type $name $version $disco(nchilds,$jid)]
#    return $text
#}

proc disco::browser::register_feature_handler {feature handler args} {
    variable browser

    set node 0
    set desc ""

    foreach {attr val} $args {
	switch -- $attr {
	    -node {set node $val}
	    -desc {set desc $val}
	}
    }

    set browser(feature_handler,$feature) $handler
    set browser(feature_handler_node,$feature) $node
    if {$desc != ""} {
	set browser(feature_handler_desc,$feature) $desc
    }
}

hook::add postload_hook \
    [list browser::register_ns_handler $::NS(disco_info) disco::browser::open_win \
	 -desc [list * [::msgcat::mc "Discover service"]]]
hook::add postload_hook \
    [list browser::register_ns_handler $::NS(disco_items) disco::browser::open_win \
	 -desc [list * [::msgcat::mc "Discover service"]]]
